# -*- coding: utf-8 -*-
# Just enough ASN.1 DER parsing to get an openssl RSA public key.
# The raw public key is just a sequence of the two integers (n, e)

from base64 import b64decode
from binascii import hexlify
from itertools import chain

__all__ = ['parse_pem', 'parse_der']

def parse_pem(pem):
    """Parse ASN.1 from an iterable yielding PEM lines."""
    der = chain.from_iterable(pem_to_bytearrays(pem))
    return parse_der(der)

def parse_der(der):
    """Parse ASN.1 from an iterable yielding DER bytes."""
    return _parse(iter(der))

def _parse(der):
    # Type tag
    ttag = next(der)
    ttype = ttag & 0x1f
    tcls = (ttag & 0b11000000) >> 6 
    constructed = bool((ttag & 0b100000) >> 5)
    
    # The variable-length length field.
    llen = next(der)
    if not llen & 0x80:
        length = llen
    else:
        length = 0
        for i in range(llen ^ 0x80):
            length <<= 8
            length += next(der)

    body = (next(der) for i in range(length))

    if ttag == 0x30: # SEQUENCE
        def seqbody():
            while True:
                yield _parse(body)
        items = list(seqbody())
        return items
    elif ttag == 0x02: # INTEGER
        return int(hexlify(bytearray(body)), 16)
    elif ttag in (0x04, 0x03): # OCTET STRING, BIT STRING
        return bytearray(body)
    elif constructed:
        return (ttag, ttype, tcls, constructed, _parse(body))
    else:
        return (ttag, ttype, tcls, constructed, bytearray(body))

def pem_to_bytearrays(pem):
    line = next(pem)
    while not line.startswith(b'-----BEGIN'):
        line = next(pem)
    line = next(pem)
    while not line.startswith(b'-----END'):
        yield bytearray(b64decode(line))
        line = next(pem)

if __name__ == "__main__": # pragma nocover
    import sys, pprint
    pprint.pprint(parse_pem(sys.stdin))
    



# -*- coding: utf-8 -*-
#
# Copyright 2013-2016 Daniel Holth <dholth@gmail.com>
# Based on python-rsa copyright 2011 Sybren A. Stüvel <sybren@stuvel.eu>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Simple pure-Python signature-verification-only RSA implementation.
"""

import base64
import binascii
import hashlib
import json
import re

__all__ = ['PublicKey', 'BadSignatureError', 'verify_jwt']

ASN1_HASH = {b'0Q0\r\x06\t`\x86H\x01e\x03\x04\x02\x03\x05\x00\x04@': 'sha512', 
             b'0A0\r\x06\t`\x86H\x01e\x03\x04\x02\x02\x05\x00\x040': 'sha384', 
             b'010\r\x06\t`\x86H\x01e\x03\x04\x02\x01\x05\x00\x04 ': 'sha256'}

# json web algorithms to Python hash functions
ALG_HASH = {'RS256': 'sha256', 'RS384': 'sha384', 'RS512': 'sha512'}

pkcs_regex = (b'\x01\xff+\x00(?P<algorithm>' +
              b'|'.join(sorted(re.escape(asn1) for asn1 in ASN1_HASH.keys())) +
              b')(?P<hash>.+)')

pkcs_signature = re.compile(pkcs_regex, re.DOTALL)

class BadSignatureError(ValueError): pass

class PublicKey(object):
    KTY = "RSA"
    
    def __init__(self, n, e, alg='RS256'):
        self.e = e
        self.n = n
        self.alg = alg
     
    @classmethod
    def from_jwk(cls, jwk):
        """Load RSA PublicKey from a JSON Web Key"""
        if jwk['kty'] != cls.KTY:
            raise ValueError("Not a {0} key".format(cls.KTY))
        n = b64_to_int(as_binary(jwk['n']))
        e = b64_to_int(as_binary(jwk['e']))        
        return cls(n, e, jwk['alg'])
    
    def to_jwk(self):
        """Return a JSON Web Key for this key."""
        jwk = {'kty':self.KTY, 
               'n':as_text(int_to_b64(self.n)),
               'e':as_text(int_to_b64(self.e)),
               'alg':self.alg}
        return jwk
    
    def verify(self, message, signature):
        """Verify a message signed with this key. Return the verified message on success.

        Input, output are bytes."""
        signature_integer = bytes_to_int(signature)
        plain_signature = decrypt(signature_integer, self)
        plain_signature_bytes = int_to_bytes(plain_signature)
        match = pkcs_signature.match(plain_signature_bytes)
        if match:
            asn1_hash = ASN1_HASH[match.group(1)]
            alg_hash = ALG_HASH[self.alg]
            if asn1_hash != alg_hash:
                raise BadSignatureError("Key and signature algorithm do not match", asn1_hash, alg_hash)
            hash = hashlib.new(ASN1_HASH[match.group(1)])
            hash.update(message)
            if hash.digest() == match.group(2):
                return message            
        raise BadSignatureError("Bad signature")
    
    def __repr__(self):
        return "{0}(n={1},e={2})".format(self.__class__.__name__, self.n, self.e)

def decrypt(ciphertext, pub):
    """RSA decryption on integers"""
    return pow(ciphertext, pub.e, pub.n)

def urlsafe_b64encode(data):
    """urlsafe_b64encode without padding"""
    return base64.urlsafe_b64encode(data).rstrip(b'=')

def urlsafe_b64decode(data):
    """urlsafe_b64decode without padding"""
    pad = b'=' * (4 - (len(data) & 3))
    return base64.urlsafe_b64decode(data + pad)

def bytes_to_int(data):
    """Convert bytes to an integer"""
    hexy = binascii.hexlify(data)
    hexy = b'0'*(len(hexy)%2) + hexy
    return int(hexy, 16)

def b64_to_int(data):
    """Convert urlsafe_b64encode(data) to an integer"""
    return bytes_to_int(urlsafe_b64decode(data))

def int_to_bytes(integer):
    hexy = as_binary('%x' % integer)
    hexy = b'0'*(len(hexy)%2) + hexy
    data = binascii.unhexlify(hexy)
    return data

def int_to_b64(integer):
    """Convert an integer to urlsafe_b64encode() data"""
    return urlsafe_b64encode(int_to_bytes(integer))

def as_binary(text):
    return text.encode('latin1')

def as_text(data):
    return data.decode('latin1')

def _verify_jwt(token, jwks):
    message, _, signature = token.rpartition('.')
    header, payload = message.split('.')
    header = json.loads(urlsafe_b64decode(header))
    kid = header['kid']
    try:
        jwk = next(key for key in jwks['keys'] if key['kid'] == kid)
    except StopIteration:
        raise BadSignatureError("Key not found in jwks")
    pk = PublicKey.from_jwk(jwk)
    verified = pk.verify(message, urlsafe_b64decode(signature))
    return verified

def verify_jwt(token, jwks):
    """Verify the signature on a RSA-signed JSON web token.
    Does not check any of the claims within the token.

    token: Full signed token
    jwks: JSON web key set (object with {"keys":[...]})

    Return decoded payload.
    """
    verified = _verify_jwt(token, jwks)
    header, _, payload = verified.partition('.')
    return json.loads(urlsafe_b64decode(payload))



import rsalette
import asn1lette
import subprocess
import json
import os.path

from nose.tools import eq_, raises

keys = [            
        {'e': 'AQAB', 
         'kty': 'RSA', 
         'alg': 'RS384',
         # 'use': 'sig',    # TODO
         'n': 'tbg-f-iI33OW7Pll0eah8mmAz-5kQWntKRzP3Bd3dB93523t0ZQEhS17wR4TLOgKBGDGhncMvWUH53-pPWxkDanPXpQ53mK4McQfA6PE__XrgUI_DfpuK-46HJecJnyKcghrSUKkKAM9ZU46zVRsmr84t8IKBwRwzdqfOT3UJEbB3ktqw-1UNsz0ZmBAeZXnETbBGwSo3tTeHOVq0E6kYqmlaO0Eu1jfN8mxLhc1x7_9osjsbO0pkJTchdaVBl7MLpmYNfwlh3eAzir__avGXetJa9fpsP2KAG0_6OSlTh2MzwyuTRTqmU0rQQOHscqoM8VubFH5odRcca3lHFFb0Q'}
       ]

pem = b"""

-----BEGIN RSA PUBLIC KEY-----
MIIBCgKCAQEAtbg+f+iI33OW7Pll0eah8mmAz+5kQWntKRzP3Bd3dB93523t0ZQE
hS17wR4TLOgKBGDGhncMvWUH53+pPWxkDanPXpQ53mK4McQfA6PE//XrgUI/Dfpu
K+46HJecJnyKcghrSUKkKAM9ZU46zVRsmr84t8IKBwRwzdqfOT3UJEbB3ktqw+1U
Nsz0ZmBAeZXnETbBGwSo3tTeHOVq0E6kYqmlaO0Eu1jfN8mxLhc1x7/9osjsbO0p
kJTchdaVBl7MLpmYNfwlh3eAzir//avGXetJa9fpsP2KAG0/6OSlTh2MzwyuTRTq
mU0rQQOHscqoM8VubFH5odRcca3lHFFb0QIDAQAB
-----END RSA PUBLIC KEY-----"""

ec_pem = b"""
-----BEGIN EC PRIVATE KEY-----
MHQCAQEEIB439n9KipyhXSmf8F9TWgOPFYj46woOmvkgUxcOeqheoAcGBSuBBAAK
oUQDQgAE5tG+pwfXBazh6HY8/Hs/j4cW0gcw485nUkoR3cgxNyAbY9tXsMAJMEIr
47LwqNpXKgufNoBL6Nmjg6WF5Bon4Q==
-----END EC PRIVATE KEY-----
"""

def test_asn1lette():
    # Do the ASN.1 keys contain an equivalent to alg?
    pk = rsalette.PublicKey(*asn1lette.parse_pem(iter(pem.splitlines())), alg='RS384')
    eq_(pk.to_jwk(), keys[0])

def test_ec_pem():
    """Not going to implement ECDSA any time soon, but we can sortof parse
    their private keys...
        
       ECPrivateKey ::= SEQUENCE {
         version        INTEGER { ecPrivkeyVer1(1) } (ecPrivkeyVer1),
         privateKey     OCTET STRING,
         parameters [0] ECParameters {{ NamedCurve }} OPTIONAL,
         publicKey  [1] BIT STRING OPTIONAL
       }
    """
    curves = { b'+\x81\x04\x00\n' : 'P-256' }
    asn1 = asn1lette.parse_pem(iter(ec_pem.splitlines()))
    assert asn1[0] == 1
    ec_key = { "kty" : "EC", "d" : asn1[1], "crv" : curves[bytes(asn1[2][-1][-1])] }

@raises(StopIteration) # ? bad exception handling
def test_asn1lette_fail():
    asn1lette.parse_der(bytearray(b'\x32'))

def test_rsalette():
    in_jwk = keys[0]
    pk = rsalette.PublicKey.from_jwk(in_jwk)
    jwk = pk.to_jwk()
    for param in ('kty', 'n', 'e'):
        eq_(jwk[param], in_jwk[param])
    
    assert 'PublicKey(' in repr(pk)
   
@raises(ValueError)
def test_bad_key():
    rsalette.PublicKey.from_jwk({'kty':'ARS'})

# Generating a suitable keypair:
#
# openssl genrsa -out test_private.pem 2048
# (the python-rsa library was used to extract the public key)

def test_signature():
    for digest in 'sha256', 'sha384', 'sha512':
        _test_signature(digest)
        
def sign_with_openssl(message, key_filename, digest='sha256'):
    """Sign message with openssl, returning the signature.
    
    message: the message in bytes
    key_filename: path to a PEM-format RSA private key
    digest: digest algorithm (see 'openssl dgst' for legal values)
    """
    command = ['openssl', 'dgst', '-' + digest, '-sign', key_filename, '-binary']
    openssl = subprocess.Popen(command, 
                               stdin=subprocess.PIPE,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE)
    output = openssl.communicate(message)
    rc = openssl.wait()
    if rc != 0:
        raise subprocess.CalledProcessError(rc, ' '.join(command), output[1])
    return output[0]

def _test_signature(digest='sha256'):
    our_pubkey = rsalette.PublicKey.from_jwk(keys[0])
    our_pubkey.alg = {'sha256':'RS256', 'sha384':'RS384', 'sha512':'RS512'}[digest]
    here = os.path.abspath(os.path.dirname(__file__))
    
    message = json.dumps(keys[0]).encode('utf-8')
    signature = sign_with_openssl(message, os.path.join(here, 'test_private.pem'), digest)
    
    verified_message = our_pubkey.verify(message, signature)
    eq_(verified_message, message)
    
    @raises(ValueError)
    def bad_signature():
        our_pubkey.verify(message+b' ', signature)
        
    bad_signature()
